from setuptools import find_packages, setup

setup(
    name='causalbench',
    packages=find_packages(include=['causalbench']),
    version='0.1.0',
    description='Spatio Temporal Causal Benchmarking Platform',
    author='Me'
)
